# STA 410, Fall 2014, Assignment #2, Second problem.


# Compute likelihood for parameters sigma and rho given data X.

bvn_likelihood <- function (X, sigma, rho)
{
    (2*pi*sigma^2*sqrt(1-rho^2))^-nrow(X) *
      exp (-sum (X[,1]^2 + X[,2]^2 - 2*rho*X[,1]*X[,2]) / (2*(1-rho^2)*sigma^2))
}


# Compute the prior probability density for sigma and rho.

bvn_prior <- function (sigma, rho)
{
    exp(-sigma) / (4*sqrt(abs(rho)))
}


# Compute the unnormalized marginal posterior distribution for rho given data X,
# by integrating over sigma.

bvn_marg_unnorm_post_rho <- function (X, rho)
{
    integrate (function (sigma) {
                   r <- numeric(length(sigma))
                   for (i in 1:length(sigma))
                       r[i] <- bvn_prior (sigma[i], rho) *
                               bvn_likelihood (X, sigma[i], rho)
                   r
               }, 0, Inf) $ value
}


# Compute the normalizing constant for the unnormalized posterior given data X.
# The required integral over rho is split into an integral from -1 to 0 and an
# integral from 0 to +1 to handle the singularity in the prior at 0.

bvn_normalize <- function (X)
{
    f <- function (rho)
    {   
        r <- numeric(length(rho))
        for (i in 1:length(rho))
            r[i] <- bvn_marg_unnorm_post_rho(X,rho[i])
        r
    }
                 
    integrate (f, -1, 0) $ value + integrate (f, 0, 1) $ value
}


# Compute the normalized posterior density for rho given data X.

bvn_posterior_rho <- function (X, rho)
{
    bvn_marg_unnorm_post_rho (X, rho) / bvn_normalize(X)
}


# Function to plot prior and posterior densities for rho given data X.

bvn_plot_prior_posterior_rho <- function (X)
{
    grid <- seq(-0.999,+0.999,length=500)

    prior <- 1 / (4*sqrt(abs(grid)))

    post <- numeric(length(grid))
    for (i in 1:length(grid)) 
        post[i] <- bvn_posterior_rho (X, grid[i])

    plot (grid, prior, type="l",ylim=c(0,5), yaxs="i", xlim=c(-1,1), xaxs="i",
          xlab="rho", ylab="prior and posterior densities", lty=3)

    lines (grid, post)
}


# Function to compute the posterior probability that rho is negative.

bvn_prob_neg_rho <- function (X)
{
    integrate (function (rho) sapply(rho, function (r) bvn_posterior_rho(X,r)),
               -1, 0) $ value
}


# Apply these functions to two artificial datasets.

set.seed(1)
n <- 40
x1 <- rnorm(n); x2 <- rnorm(n); z <- rnorm(n)
Xa <- cbind (x1, x2); Xb <- cbind (x1+0.5*z, x2+0.5*z)

pdf("ass2bvn.pdf",width=8,height=6,pointsize=10)
par(mfrow=c(1,2))

bvn_plot_prior_posterior_rho (Xa)
abline (v = cor(Xa[,1],Xa[,2]), lty=3)
abline (v = 0)
title("Dataset Xa")

cat("Posterior probability of negative rho for Xa:",bvn_prob_neg_rho(Xa),"\n")

bvn_plot_prior_posterior_rho (Xb)
abline (v = cor(Xb[,1],Xb[,2]), lty=3)
abline (v = 0.2)
title("Dataset Xb")

cat("Posterior probability of negative rho for Xb:",bvn_prob_neg_rho(Xb),"\n")


# See for what size dataset (with the same true distribution as Xb) underflow 
# becomes a problem.

for (n in 41:60) {
    set.seed(1)
    x1 <- rnorm(n); x2 <- rnorm(n); z <- rnorm(n)
    X <- cbind (x1+0.5*z, x2+0.5*z)
    cat ("\nn =", n, "\n")
    cat("  normalization constant:", bvn_normalize(X), "\n")
}

dev.off()
